#ifndef _HKXCONTROTATION_H
#define _HKXCONTROTATION_H

/**
 * \file HkxContRotation.h
 * \brief Servo continuous rotation of the Arduino library for HerkuleX servo (DRS-0101 and DRS-0201 models)
 * \author Benoit Puel (contact@benoitpuel.com)
 * \version 1.2
 * \date 27 February 2016
 * \copyright This project is released under the GNU GPL V3.
 */

#include "Arduino.h"
#include "HkxCommunication.h"
#include "HkxStatus.h"

/**
 * \class HkxContRotation
 * \brief Class to drive one servo in continuous rotation mode.
 * \details This class allows to drive one servo in continuous rotation mode.
 */
class HkxContRotation{
private:
  /** Communication with servos  */
  HkxCommunication& _herkXCom; 
  /** Communication to print messages   */
  HkxPrint& _print; 
  /** Connexion with the servo to drive  */
  boolean _connected; 
  /** \cond developer */
  /** Class name to print the messages */
  const String _className;
  /** \endcond */
  /** ID of the servo  */
  uint8_t _id;
  /** Current RAM value of the torque control of the servo  */
  hkxTorqueControl _torqueControl;
  /** Current RAM value of the LED of the servo  */
  hkxLEDControl _ledControl;
  /** Last status received from the servo  */
  HkxStatus _statusED;

  /** \cond developer */  
  /**
   * \brief Update the status
   * \details Perform a status request to update the stored status \ref _statusED.
   * \return 
   *          0 = OK \n
   *          1 = Input not correct \n
   *          2 = Servo not connected \n
   *          3 = Data not consistent
   */
  uint8_t updateStatus();

  /**
   * \brief Print error
   * \details Print an error message if setup.
   * \param[in] message : content to print as an error.  
   */
  void errorPrint(const String& message){if(_print._errorMessages){_print.errorPrint(_className + message);}};

  /**
   * \brief Print warning
   * \details Print a warning message if setup.
   * \param[in] message : content to print as a warning.  
   */
  void warningPrint(const String& message){if(_print._errorMessages){_print.warningPrint(_className + message);}};

  /**
   * \brief Print info
   * \details Print an info message if setup.
   * \param[in] message : content to print as an info.  
   */
  void infoPrint(const String& message){if(_print._errorMessages){_print.infoPrint(_className + message);}};
  /** \endcond */

public:  
  /**
   * \brief Constructor
   * \details The constructor of HkxContRotation.
   * \param[in] ID : id of the servo to drive. The value shall be [0 ; 253].
   * \param[in] herkXCom : Communication with the servos. 
   * \param[in] print : Communication to print messages.
   *
   * Example:
   * \code 
   * HkxPrint print = HkxPrint(Serial, 9600); 
   * HkxCommunication communication = HkxCommunication(HKX_115200, Serial1, print); 
   * HkxContRotation servo = HkxContRotation(10, communication, print); // ID=10, print can be different to communication
   * \endcode
   */
  HkxContRotation(uint8_t ID, HkxCommunication& herkXCom, HkxPrint& print);
  
  /** \cond developer */
  /**
   * \brief Try connect servo
   * \details Try to connect to the servo.
   * \return
   * - \c true: if the servo is connected, \n
   * - \c false: if the servo is NOT connected. 
   */
  boolean tryConnect();
  /** \endcond */

  /**
   * \brief Get the ID
   * \details Get the ID of the driven servo.
   * \return Return the ID. 
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * ...
   * byte idOfMyServo = servo.getID();
   * \endcode
   */
  uint8_t getID() const {return this->_id;}; 

  /**
   * \brief Get torque control
   * \details Get the current torque control of the driven servo (see \ref hkxTorqueControl for more details about the options).
   * \return Return the torque control 
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * hkxTorqueControl torqueControlOfMyServo = servo.getTorqueControl();
   * \endcode
   */
  hkxTorqueControl getTorqueControl() const {return this->_torqueControl;};

  /**
   * \brief Get LED control
   * \details Get the current LED control (colour) of the driven servo (see \ref hkxLEDControl for more details about the options).
   * \return Return the LED control. 
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * hkxLEDControl ledControlOfMyServo = servo.getLEDControl();
   * \endcode
   */
  hkxLEDControl getLEDControl() const {return this->_ledControl;};

  /** \cond developer */
  /**
   * \brief Get servo communication
   * \details Get the communication of the servo.
   * \return Return the communication.
   */
  HkxCommunication& getCommunication() const {return this->_herkXCom;};

  /**
   * \brief Is the servo connected
   * \details Return if the servo is connected.
   * \return 
   * - \c true: if the servo is connected (last communication worked), \n
   * - \c false: if the servo is NOT connected (last communication failed). 
   */
  boolean isConnected() const {return this->_connected;};

  /**
   * \brief Set the LED variable
   * \details Set the LED variable in the class without applying. 
   * \warning This member is for \ref HkxGroupPosControl use only.
   * \param[in] newLEDControl : Colour of the led (see \ref hkxLEDControl for more details).
   */
  void setLEDControlVariable(hkxLEDControl newLEDControl){this->_ledControl = newLEDControl;};
  /** \endcond */

  /**
   * \brief Check if the servo stopped
   * \details Check if the servo is stopped.
   * \return 
   * - \c true if the servo is stopped.
   * - \c false if the servo is still moving.
   *
   * Example:
   * \code 
   * ... 
   * HkxPosControl servo = HkxPosControl(10, communication, print);
   * servo.moveRotation(512, 2000, HKX_NO_VALUE); // move at 512 PWM during 2 seconds
   * ... // different actions done during the movement
   * while(!servo.isStopped()){} // wait the servo to be stopped
   * \endcode
   */
  boolean isStopped(){
    this->updateStatus();
    return !this->_statusED.isDetail(HKX_STAT_MOVING_FLAG);
  };

  /**
   * \brief Get the status
   * \details Get the last received status from the servo.
   * \param[out] statusED : Last received status of the servo.
   * \param[out] update : 
   * - \c true to update the status, then to get the current status.
   * - \c false to get the last received status.
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected \n
   * 3 = Data not consistent
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * HkxStatus currentStatus;
   * byte error = servo.getStatus(currentStatus, true);
   * if(error==0 && currentStatus.isError(HKX_STAT_VOLTAGE)){ // check voltage error
   *   print.errorPrint(F("Please check input voltage")); 
   * }
   * \endcode
   */
  uint8_t getStatus(HkxStatus& statusED, boolean update);

  /**
   * \brief Clear the status
   * \details Clear the status of the servo to leave the "error state" and stop the LED blink and torque deactivation (see p.33 of the user manual for more details about LED and torque policies).
   * \warning If the problem still present (e.g. low input voltage), the servo will switch again to error just after clearing. 
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * HkxStatus currentStatus;
   * byte error = servo.getStatus(currentStatus, true);
   * if(error==0 && currentStatus.isError(HKX_STAT_ALL)){ // check if any error
   *   servo.clearStatus(); 
   * }
   * \endcode
   */
  uint8_t clearStatus();

  /**
   * \brief Set torque actuation and LED colour
   * \details Set the torque actuation (\ref hkxTorqueControl for more details) and the LED colour (\ref hkxLEDControl for more details). 
   * \param[in] newTorqueControl : Torque control to apply. This parameter is optional, either set a \ref hkxTorqueControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_TORQUE_FREE: no resistance to the movements (or let say mechanical resistance only)
   * - \c HKX_TORQUE_BREAK: the motor resists but allows the movements
   * - \c HKX_TORQUE_ON: the motor maintain its position (don't allow movements)
   * \param[in] newLEDControl : LED control to apply. This parameter is optional, either set a \ref hkxLEDControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_LED_OFF,
   * - \c HKX_LED_GREEN,
   * - \c HKX_LED_BLUE,
   * - \c HKX_LED_RED,
   * - \c HKX_LED_CYAN,
   * - \c HKX_LED_YELLOW,
   * - \c HKX_LED_PINK,
   * - \c HKX_LED_WHITE,
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * servo.setTorqueLEDControl(HKX_LED_GREEN, HKX_NO_VALUE); // set the led to green
   * servo.setTorqueLEDControl(HKX_LED_WHITE, HKX_TORQUE_BREAK); // set the led to white and the torque to break 
   * \endcode
   */
  uint8_t setTorqueLEDControl(HkxMaybe<hkxTorqueControl> newTorqueControl, HkxMaybe<hkxLEDControl> newLEDControl);

  /**
   * \brief Move continuous rotation
   * \details Move the servo in continuous rotation.
   * \param[in] pwm : Torque power to apply. Its value shall be set in PWM within the range [-1023 ; 1023]. 
   * \param[in] playTime : Time of the action. Its value shall be set in milliseconds. This parameter is optional, either set a \c uint16_t (or \c unsigned \c int) variable, or set \c HKX_NO_VALUE to ignore it.
   * \param[in] LEDControl : LED control to apply. This parameter is optional, either set a \ref hkxLEDControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_LED_OFF,
   * - \c HKX_LED_GREEN,
   * - \c HKX_LED_BLUE,
   * - \c HKX_LED_RED,
   * - \c HKX_LED_CYAN,
   * - \c HKX_LED_YELLOW,
   * - \c HKX_LED_PINK,
   * - \c HKX_LED_WHITE,
move.
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   *
   * Example:
   * \code 
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * servo.move(1200, 1500, HKX_NO_VALUE, true); // move to position 120° in 1.5 seconds and wait arrival 
   * \endcode
   */
  uint8_t moveRotation(int16_t pwm, HkxMaybe<uint16_t> playTime, HkxMaybe<hkxLEDControl> LEDControl);

  /**
   * \brief Get the servo behaviour
   * \details Get the current servo behaviour. See the user manual for more details about the parameters.
   * \param[out] inputVoltage : Current servo input voltage. Its value is given in millivolts. This parameter is optional, either set the address of a \c uint16_t (or \c unsigned \c int) variable to return the value, or set \c HKX_NO_VALUE to ignore it.
   * \param[out] temperature : Current temperature of the servo. Its value is given in 10^-2 °C (degree Celsius). This parameter is optional, either set the address of a \c uint16_t (or \c unsigned \c int) variable to return the value, or set \c HKX_NO_VALUE to ignore it.
   * \param[out] velocity : Current angle velocity of the servo. Its value is given in degree / second. This parameter is optional, either set the address of a \c uint16_t (or \c unsigned \c int) variable to return the value, or set \c HKX_NO_VALUE to ignore it.
   * \param[out] PWM : Current PWM (load) of the servo. Its value is given in PWM (raw). This parameter is optional, either set the address of a \c int16_t (or \c int) variable to return the value, or set \c HKX_NO_VALUE to ignore it.
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   *
   * Example 1: How to get all the values.
   * \code
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * uint16_t inputVoltage, temperature;
   * int16_t velocity, PWM;
   * servo->getBehaviour(&inputVoltage, &temperature, &velocity, &PWM);
   * \endcode
   * Example 2: How to get the current velocity only.
   * \code
   * int16_t velocity;
   * servo->getBehaviour(HKX_NO_VALUE, HKX_NO_VALUE, &velocity, HKX_NO_VALUE);
   * \endcode
   */
  uint8_t getBehaviour(HkxMaybe<uint16_t> inputVoltage, HkxMaybe<uint16_t> temperature, HkxMaybe<int16_t> velocity, HkxMaybe<int16_t> PWM);

  /**
   * \brief Reboot servo
   * \details Reboot the servo. This action will erase the ram, then reinitialize the parameters with the EEP (ROM) values. Wait some time (> 500 ms) before asking any action from the servo.
   * \return 
   * 0 = OK \n
   * 2 = Servo not connected
   *
   * Example:
   * \code
   * ... 
   * HkxContRotation servo = HkxContRotation(10, communication, print);
   * servo->reboot();
   * wait(500); // wait for the servo to restart
   * ...
   * \endcode
   */
  uint8_t reboot();
};

#endif    // _HKXCONTROTATION_H



