#ifndef _HKXGROUPCONTROTATION_H
#define _HKXGROUPCONTROTATION_H

/**
 * \file HkxGroupContRotation.h
 * \brief Group of servos for continuous rotation of the Arduino library for HerkuleX servo (DRS-0101 and DRS-0201 models)
 * \author Benoit Puel (contact@benoitpuel.com)
 * \version 1.2
 * \date 27 February 2016
 * \copyright This project is released under the GNU GPL V3.
 */

#include "HkxContRotation.h"

/**
 * \class HkxGroupContRotation
 * \brief Class to drive a group of servos in continuous rotation mode.
 * \details This class allows to drive a group of servos in continuous rotation mode.
 */
class HkxGroupContRotation{
private:
  /** Array of pointers to \ref HkxContRotation instances  */
  HkxContRotation** _servos;
  /** Number of servos in the table \ref _servos  */
  const uint8_t _length;
  /** Communication with servos  */
  HkxCommunication& _herkXCom;
  /** Communication to print messages  */
  HkxPrint& _print; 
  /** \cond developer */
  /** Class name to print the messages */
  const String _className;


  /**
   * \brief Print error
   * \details Print an error message if setup.
   * \param[in] message : content to print as an error.  
   */
  void errorPrint(const String& message){if(_print._errorMessages){_print.errorPrint(_className + message);}};

  /**
   * \brief Print warning
   * \details Print a warning message if setup.
   * \param[in] message : content to print as a warning.  
   */
  void warningPrint(const String& message){if(_print._errorMessages){_print.warningPrint(_className + message);}};

  /**
   * \brief Print info
   * \details Print an info message if setup.
   * \param[in] message : content to print as an info.  
   */
  void infoPrint(const String& message){if(_print._errorMessages){_print.infoPrint(_className + message);}};
  /** \endcond */

public:  
  /**
   * \brief Constructor
   * \details The constructor of HkxGroupContRotation.
   * \warning \ref HkxCommunication shall be the same for all the servos of the group. If not, an error will be returned and the instance will be corrupted.
   * \param[in] length : Number of servos in the table \ref _servos. 
   * \warning It shall not exceed 53 servos, otherwise it cannot execute \ref moveAllRotation(). 
   * \param[in] arrayServos[] : Array of pointers to the servos. 
   * \param[in] print : Communication to print messages.
   */
  HkxGroupContRotation(uint8_t length, HkxContRotation* arrayServos[], HkxPrint& print);
  
  /**
   * \brief Get the nth servo
   * \details Get the nth servo.
   * \param[in] number : number of servo to get (from the table \ref _servos). 
   * \warning \c number < _ref _length otherwise return void.
   * \return Return a reference to the nth servo. 
   */
  HkxContRotation& getServo(uint8_t number){if(number < _length){return *_servos[number];}};

  /**
   * \brief Get number of servos
   * \details Get the number of servos manage by the instance.
   * \return Return the number of servos. 
   */
  uint8_t getNbServos(){return this->_length;};

  /**
   * \brief Set torque actuation and LED colour
   * \details Set the same torque actuation (\ref hkxTorqueControl for more details) and the LED colour (\ref hkxLEDControl for more details) to all the servos. 
   * \param[in] newTorqueControl : Torque control to apply. This parameter is optional, either set a \ref hkxTorqueControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_TORQUE_FREE: no resistance to the movements (or let say mechanical resistance only)
   * - \c HKX_TORQUE_BREAK: the motor resists but allows the movements
   * - \c HKX_TORQUE_ON: the motor maintain its position (don't allow movements)
   * \param[in] newLEDControl : LED control to apply. This parameter is optional, either set a \ref hkxLEDControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_LED_OFF,
   * - \c HKX_LED_GREEN,
   * - \c HKX_LED_BLUE,
   * - \c HKX_LED_RED,
   * - \c HKX_LED_CYAN,
   * - \c HKX_LED_YELLOW,
   * - \c HKX_LED_PINK,
   * - \c HKX_LED_WHITE,
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   */
  uint8_t setAllTorqueLEDControl(HkxMaybe<hkxTorqueControl> newTorqueControl, HkxMaybe<hkxLEDControl> newLEDControl);

  /**
   * \brief Move servos in continuous rotation 
   * \details Perform a move of all the servos in continuous rotation mode.
   * \param[in] pwm[] : Array of torque power to apply. Its value shall be set in PWM within the range [-1023 ; 1023]. 
   * \param[in] playTime : Time of the action. Its value shall be set in milliseconds. This parameter is optional, either set a \c uint16_t (or \c unsigned \c int) variable, or set \c HKX_NO_VALUE to ignore it.
   * \param[in] LEDControl[] : Array of LED control to apply. This parameter is optional, either set a \ref hkxLEDControl variable, the value (list below), or set \c HKX_NO_VALUE to ignore it.
   * - \c HKX_LED_OFF,
   * - \c HKX_LED_GREEN,
   * - \c HKX_LED_BLUE,
   * - \c HKX_LED_RED,
   * - \c HKX_LED_CYAN,
   * - \c HKX_LED_YELLOW,
   * - \c HKX_LED_PINK,
   * - \c HKX_LED_WHITE,
   * \return 
   * 0 = OK \n
   * 1 = Input not correct \n
   * 2 = Servo not connected
   */
  uint8_t moveAllRotation(int16_t pwm[], HkxMaybe<uint16_t> playTime, HkxMaybe<hkxLEDControl> LEDControl[]); 

  /**
   * \brief Reboot 
   * \details Reboot all the servos.
   */
  void rebootAll(){
    for(int i=0 ; i < this->_length ; i++){
      this->_servos[i]->reboot();
    }
  };
};

#endif    // _HKXGROUPCONTROTATION_H

